use std::ops::Add;

use tiny_skia::IntRect;

use super::{cache::NodeState, LayoutNodeAccess, SizePolicy};

pub trait ArrangementCalculator {
    fn arrange_step(
        &self,
        node: LayoutNodeAccess,
        child_policies: Vec<(SizePolicy, SizePolicy)>,
    ) -> (SizePolicy, SizePolicy);
    fn layout_step(&self, node: LayoutNodeAccess, inside: IntRect);
}

#[derive(Clone, Debug)]
pub enum LineArrangement {
    Column,
    Row,
}

impl LineArrangement {
    fn is_column(&self) -> bool {
        match self {
            Self::Column => true,
            Self::Row => false,
        }
    }
}

impl ArrangementCalculator for LineArrangement {
    fn arrange_step(
        &self,
        node: LayoutNodeAccess,
        child_policies: Vec<(SizePolicy, SizePolicy)>,
    ) -> (SizePolicy, SizePolicy) {
        if child_policies.is_empty() {
            return (node.width_policy, node.height_policy);
        }

        let cw_policy = child_policies
            .iter()
            .map(|v| v.0)
            .reduce(if self.is_column() {
                SizePolicy::max
            } else {
                SizePolicy::add
            })
            .unwrap();
        let ch_policy: SizePolicy = child_policies
            .iter()
            .map(|v| v.1)
            .reduce(if self.is_column() {
                SizePolicy::add
            } else {
                SizePolicy::max
            })
            .unwrap();
        (
            node.width_policy.max_preserve_slack(cw_policy),
            node.height_policy.max_preserve_slack(ch_policy),
        )
    }

    fn layout_step(&self, node: LayoutNodeAccess, inside: IntRect) {
        // do the final children layouts
        node.cache
            .with_state(node.cache_key, |ns| ns.rect = Some(inside));

        if node.child_len() == 0 {
            return;
        }

        // expansion direction extraction lambda
        let ee = if self.is_column() {
            |ns: &mut NodeState| ns.net_policy.1
        } else {
            |ns: &mut NodeState| ns.net_policy.0
        };

        let policies = node
            .child_iter()
            .map(|c| node.cache.with_state(c.cache_key, ee).unwrap());
        let mut last_offset = 0;
        let fit = do_fit(
            if self.is_column() {
                inside.height() as i32
            } else {
                inside.width() as i32
            },
            policies,
        );
        for (offset, child) in fit.zip(node.child_iter()) {
            let crect = if self.is_column() {
                IntRect::from_xywh(
                    inside.left(),
                    inside.top() + last_offset,
                    inside.width(),
                    (offset - last_offset) as u32,
                )
                .unwrap()
            } else {
                IntRect::from_xywh(
                    inside.left() + last_offset,
                    inside.top(),
                    (offset - last_offset) as u32,
                    inside.height(),
                )
                .unwrap()
            };

            self.layout_step(child, crect);
            last_offset = offset;
        }
    }
}

fn do_fit<'l>(
    total: i32,
    policies: impl Iterator<Item = SizePolicy> + Clone,
) -> impl Iterator<Item = i32> {
    // first pass over children: collect sum total of minimum/desired sizes and/slack weights
    let policy_sum = policies.clone().reduce(SizePolicy::add).unwrap_or_default();

    if (policy_sum.desired - policy_sum.minimum) == 0 {}

    // how much of the desired size can we fit?
    let fit_coeff = (total - policy_sum.minimum as i32) as f32
        / (policy_sum.desired - policy_sum.minimum) as f32;
    let fit_coeff = if fit_coeff.is_nan() {
        1.0
    } else {
        // not more than 100% of desired
        fit_coeff.min(1.0)
    };

    // how much slack space do we have left?
    let slack_coeff = if policy_sum.slack_weight > 0 && fit_coeff >= 1.0 {
        (total - policy_sum.desired as i32) as f32 / policy_sum.slack_weight as f32
    } else {
        0.0
    };

    let mut offset = 0;
    // second pass over children: actually space out according to size policies and calculated coefficients
    policies.map(move |policy| {
        let mut amount = policy.minimum as f32;
        amount += (policy.desired - policy.minimum) as f32 * fit_coeff;
        amount += policy.slack_weight as f32 * slack_coeff;
        let amount = amount.round() as i32;
        offset += amount;
        offset
    })
}
