use std::cell::RefCell;
use std::rc::Rc;

use tiny_skia::IntRect;

use crate::input::InputState;
use crate::layout::{self, LayoutNode, LayoutNodeAccess, LinearAccess};
use crate::widget::Widget;
use crate::{component::Component, ui::UIHandle};

#[derive(Debug)]
pub enum WindowEvent {
    CloseRequested,
    Resized,
}

pub trait WindowComponent: 'static + Sized + Component {
    fn map_window_event(&self, we: WindowEvent) -> Option<Self::Msg>;

    type RootWidget: Widget<Self>;
    fn root_widget(&self) -> &Self::RootWidget;
    fn root_widget_mut(&mut self) -> &mut Self::RootWidget;
}

pub(crate) struct WindowState<WC: WindowComponent> {
    wc: WC,
    root_node: LayoutNode,
    window: Rc<winit::window::Window>,
    events: Vec<WindowEvent>,
    istate: InputState,
    draw_pending: bool,
    surface: softbuffer::Surface<Rc<winit::window::Window>, Rc<winit::window::Window>>,
}

impl<WC: WindowComponent> WindowState<WC> {
    fn redraw(&mut self) {
        self.draw_pending = false;

        let size = self.window.inner_size();
        self.surface
            .resize(
                size.width.try_into().unwrap(),
                size.height.try_into().unwrap(),
            )
            .unwrap();
        let mut buf = self.surface.buffer_mut().unwrap();

        self.root_node.set_behaviour(layout::NodeBehaviour::Fixed {
            rect: IntRect::from_xywh(0, 0, size.width, size.height).unwrap(),
        });

        let layout = LinearAccess::new(&self.root_node, self.wc.root_widget().layout_node());

        let mut pixmap = tiny_skia::Pixmap::new(size.width, size.height).unwrap();
        layout::recalculate(LayoutNodeAccess::new(&layout));
        self.wc.root_widget().render(&mut pixmap, &());

        // expensive ABGR -> RGB0 conversion
        let mut pxdata = pixmap.data();
        for t in buf.iter_mut() {
            *t = u32::from_be_bytes(pxdata[0..4].try_into().unwrap()) >> 8; //((pxdata[0] as u32) << 16) | ((pxdata[1] as u32) << 8) | (pxdata[2] as u32);
            pxdata = &pxdata[4..];
        }

        buf.present().unwrap();
    }

    fn poll(&mut self) -> Vec<WC::ParentMsg> {
        let mut ret = vec![];
        for we in self.events.drain(..) {
            let mapped = self.wc.map_window_event(we);
            ret.extend(self.wc.process_all(mapped.into_iter()).into_iter());
        }

        let evts = self.wc.root_widget_mut().poll(Some(&self.istate));
        ret.extend(self.wc.process_all(evts.into_iter()));

        if self.draw_pending {
            self.redraw();
        }

        ret
    }
}

pub(crate) trait WindowStateAccess {
    fn push_event(&self, we: WindowEvent);
    fn request_redraw(&self);
}

impl<WC: WindowComponent> WindowStateAccess for RefCell<WindowState<WC>> {
    fn push_event(&self, we: WindowEvent) {
        self.borrow_mut().events.push(we);
    }
    fn request_redraw(&self) {
        self.borrow_mut().draw_pending = true;
    }
}

pub struct WindowBuilder<'r, 'l: 'r> {
    ui_handle: &'r mut UIHandle<'l>,
}

impl<'r, 'l: 'r> WindowBuilder<'r, 'l> {
    pub(crate) fn new(ui_handle: &'r mut UIHandle<'l>) -> Self {
        Self { ui_handle }
    }

    pub fn build<WC: WindowComponent>(self, wc: impl FnOnce(&UIHandle) -> WC) -> Window<WC> {
        let window = Rc::new(
            self.ui_handle
                .eloop
                .create_window(winit::window::WindowAttributes::default())
                .unwrap(),
        );
        let wid = window.id();

        let ctx = softbuffer::Context::new(window.clone()).unwrap();
        let surface = softbuffer::Surface::new(&ctx, window.clone()).unwrap();

        let wc = wc(&self.ui_handle);
        let wstate = Rc::new(RefCell::new(WindowState {
            wc,
            root_node: LayoutNode::new(self.ui_handle.state.layout_cache.clone()),
            window,
            events: Default::default(),
            istate: InputState::default().into(),
            draw_pending: false.into(),
            surface: surface.into(),
        }));

        self.ui_handle.state.window_states.insert(
            wid,
            Rc::downgrade(&(wstate.clone() as Rc<dyn WindowStateAccess>)),
        );

        Window { state: wstate }
    }
}

pub struct Window<WC: WindowComponent> {
    state: Rc<RefCell<WindowState<WC>>>,
}

impl<WC: WindowComponent> Window<WC> {
    pub fn poll(&mut self) -> Vec<WC::ParentMsg> {
        self.state.borrow_mut().poll()
    }
}
